package me.unurled.sacredrealms.sr.commands.admin;

import static me.unurled.sacredrealms.sr.utils.Logger.log;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import me.unurled.sacredrealms.sr.components.clientbuild.ClientBuild;
import me.unurled.sacredrealms.sr.components.clientbuild.ClientBuildManager;
import org.bukkit.Location;
import org.bukkit.block.Block;
import org.bukkit.block.data.BlockData;
import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.command.TabExecutor;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class ClientBuildCommand implements TabExecutor, Listener {

  private List<String> nameCompletions = new ArrayList<>();
  private ClientBuildManager manager = null;

  /**
   * Executes the given command, returning its success. <br>
   * If false is returned, then the "usage" plugin.yml entry for this command (if defined) will be
   * sent to the player.
   *
   * @param sender Source of the command
   * @param command Command which was executed
   * @param label Alias of the command which was used
   * @param args Passed command arguments
   * @return true if a valid command, otherwise false
   */
  @Override
  public boolean onCommand(
      @NotNull CommandSender sender,
      @NotNull Command command,
      @NotNull String label,
      @NotNull String[] args) {
    if (!sender.hasPermission("sr.clientbuild")) {
      sender.sendMessage("You do not have permission to use this command.");
      return true;
    }

    if (!(sender instanceof Player p)) {
      sender.sendMessage("You must be a player to use this command.");
      return true;
    }

    if (args.length == 0) {
      sender.sendMessage("Usage: /clientbuild <create|delete|list|modify|save|display>");
      return true;
    }

    ClientBuildManager manager = ClientBuildManager.getInstance(ClientBuildManager.class);
    if (args.length == 1) { // list
      if (args[0].equalsIgnoreCase("list")) {

        List<ClientBuild> builds = manager.getBuilds();
        if (builds.isEmpty()) {
          sender.sendMessage("There are no client builds.");
          return true;
        }

        sender.sendMessage("Client Builds:");
        for (ClientBuild build : builds) {
          if (build != null) sender.sendMessage(build.getName());
        }
        return true;

      } else {
        sender.sendMessage("Usage: /clientbuild <create|delete|modify|save|display> <name>");
        return true;
      }
    } else if (args.length == 2) {
      // create, delete, modify, save, display
      if (args[0].equalsIgnoreCase("create")) {
        String name = args[1].toLowerCase();
        if (manager.getBuildNames().contains(name)) {
          sender.sendMessage("A client build with that name already exists.");
          return true;
        }
        sender.sendMessage(
            "Client build created, place blocks that you want to add to the Client"
                + " Build and save after you are finished.");
        manager.setTemporaryBlocks(p, new HashMap<>());
        manager.setTemporaryPreviousBlocks(p, new HashMap<>());
        return true;
      } else if (args[0].equalsIgnoreCase("delete")) {
        String name = args[1].toLowerCase();
        if (!manager.getBuildNames().contains(name)) {
          sender.sendMessage("A client build with that name does not exist.");
          return true;
        }
        manager.removeBuild(name);
        sender.sendMessage("Client build deleted.");
        return true;
      } else if (args[0].equalsIgnoreCase("modify")) {
        String name = args[1].toLowerCase();
        if (!manager.getBuildNames().contains(name)) {
          sender.sendMessage("A client build with that name does not exist.");
          return true;
        }
        sender.sendMessage("Modify the client build and save after you are finished.");
        manager.setTemporaryBlocks(p, new HashMap<>());
        manager.setTemporaryPreviousBlocks(p, new HashMap<>());
        return true;
      } else if (args[0].equalsIgnoreCase("save")) {
        // save the client build to the list of builds so stopping the listener.
        ClientBuild build = new ClientBuild(args[1].toLowerCase());
        if (!manager.hasTemporaryBlocks(p)) {
          sender.sendMessage("You do not have any temporary blocks to save.");
          return true;
        }
        HashMap<Location, BlockData> blocks = manager.getTemporaryBlocks(p);
        log("blocks: " + blocks.size());
        for (Entry<Location, BlockData> block : blocks.entrySet()) {
          build.addBlock(block.getKey(), block.getValue());
        }

        manager.addBuild(build);

        // remove blocks from the world
        for (Entry<Location, BlockData> block : manager.getTemporaryPreviousBlocks(p).entrySet()) {
          block.getKey().getWorld().setBlockData(block.getKey(), block.getValue());
          //          p.sendBlockChange(block.getLocation(), block.getBlockData());
        }

        manager.removeTemporaryBlocks(p);
        manager.removeTemporaryPreviousBlocks(p);

        sender.sendMessage("Client builds saved.");
        return true;
      } else if (args[0].equalsIgnoreCase("display")) {
        String name = args[1].toLowerCase();
        if (!manager.getBuildNames().contains(name)) {
          sender.sendMessage("A client build with that name does not exist.");
          return true;
        }
        ClientBuild build = manager.getBuild(name);
        if (build == null) {
          sender.sendMessage("An error occurred while displaying the client build.");
          return true;
        }
        if (manager.getPlayerBuilds(p).contains(name)) {
          manager.removeDisplayFromPlayer(p, build);
          sender.sendMessage("Removed the client build display.");
        } else {
          manager.displayToPlayer(p, build);
          sender.sendMessage("Displayed the client build.");
        }
        return true;
      } else {
        sender.sendMessage("Usage: /clientbuild <create|delete|modify|save|display> <name>");
        return true;
      }
    } else {
      sender.sendMessage("Usage: /clientbuild <create|delete|modify|save|display> <name>");
      return true;
    }
  }

  @EventHandler
  public void onBlockPlace(@NotNull BlockPlaceEvent e) {
    Player p = e.getPlayer();
    if (manager == null) {
      manager = ClientBuildManager.getInstance(ClientBuildManager.class);
    }
    if (manager.hasTemporaryBlocks(p)) {
      log("block placed " + e.getBlock().getType());
      manager.getTemporaryBlocks(p).put(e.getBlock().getLocation(), e.getBlock().getBlockData());
    }
    if (manager.hasTemporaryPreviousBlocks(p)) {
      Block block = e.getBlock();
      // set blockstate to block
      manager
          .getTemporaryPreviousBlocks(p)
          .put(block.getLocation(), e.getBlockReplacedState().getBlockData());
    }
  }

  @EventHandler
  public void onBlockBreak(@NotNull BlockBreakEvent e) {
    Player p = e.getPlayer();
    if (manager == null) {
      manager = ClientBuildManager.getInstance(ClientBuildManager.class);
    }
    if (manager.hasTemporaryBlocks(p)) {
      manager.getTemporaryBlocks(p).remove(e.getBlock().getLocation());
    }
    if (manager.hasTemporaryPreviousBlocks(p)) {
      Block block = e.getBlock();
      // set blockstate to block
      manager.getTemporaryPreviousBlocks(p).put(block.getLocation(), block.getBlockData());
    }
  }

  /**
   * Requests a list of possible completions for a command argument.
   *
   * @param sender Source of the command. For players tab-completing a command inside of a command
   *     block, this will be the player, not the command block.
   * @param command Command which was executed
   * @param label Alias of the command which was used
   * @param args The arguments passed to the command, including final partial argument to be
   *     completed
   * @return A List of possible completions for the final argument, or null to default to the
   *     command executor
   */
  @Override
  public @Nullable List<String> onTabComplete(
      @NotNull CommandSender sender,
      @NotNull Command command,
      @NotNull String label,
      @NotNull String[] args) {
    if (args.length == 1) {
      return List.of("create", "delete", "list", "modify", "save", "display");
    } else if (args.length == 2) {
      if (args[0].equalsIgnoreCase("delete")
          || args[0].equalsIgnoreCase("modify")
          || args[0].equalsIgnoreCase("display")) {
        ClientBuildManager manager = ClientBuildManager.getInstance(ClientBuildManager.class);
        List<String> buildNames = manager.getBuildNames();
        if (nameCompletions.isEmpty() || !new HashSet<>(nameCompletions).containsAll(buildNames)) {
          nameCompletions = buildNames;
        }
        return nameCompletions;
      }
    }
    return null;
  }
}
